from tests.unit_tests.utils import BaseTest


class Test_HideVersion(BaseTest):
    def test_typical(self, user, moderator):

        self.login_user(moderator)

        doc_v1 = self.create_document(data="v1").json["document"]
        doc_v2 = self.modify_document(doc_v1, data="v2").json["document"]

        self.hide_version(doc_v1)

        changes = self.get_versions(document=doc_v1).json["versions"]
        assert changes[1]["version_id"] == doc_v1["version_id"]
        assert changes[1]["hidden"] is True
        assert changes[1]["data"] == "v1"
        assert changes[0]["version_id"] == doc_v2["version_id"]
        assert changes[0]["hidden"] is False
        assert changes[0]["data"] == "v2"

        self.logout_user()

        changes = self.get_versions(document=doc_v1).json["versions"]
        assert changes[1]["hidden"] is True
        assert "data" not in changes[1]

        assert changes[0]["hidden"] is False
        assert changes[0]["data"] == "v2"

        self.login_user(user)
        changes = self.get_versions(document=doc_v1).json["versions"]
        assert changes[1]["hidden"] is True
        assert "data" not in changes[1]
        assert changes[0]["hidden"] is False
        assert changes[0]["data"] == "v2"
        self.logout_user()

        self.login_user(moderator)
        self.unhide_version(doc_v1)

        changes = self.get_versions(document=doc_v1).json["versions"]
        assert changes[1]["hidden"] is False
        assert changes[1]["data"] == "v1"
        assert changes[0]["hidden"] is False
        assert changes[0]["data"] == "v2"

        self.logout_user()

        changes = self.get_versions(document=doc_v1).json["versions"]
        assert changes[1]["hidden"] is False
        assert changes[1]["data"] == "v1"
        assert changes[0]["hidden"] is False
        assert changes[0]["data"] == "v2"

    def test_forbidden(self, user, admin):

        # anonymous
        self.hide_version(1, expected_status=403)
        self.unhide_version(1, expected_status=403)

        # authenticated
        self.login_user(user)
        self.hide_version(1, expected_status=403)
        self.unhide_version(1, expected_status=403)
        self.logout_user()

        self.login_user(admin)
        self.hide_version(1, expected_status=403)
        self.unhide_version(1, expected_status=403)

    def test_notfound(self, moderator):
        self.login_user(moderator)
        self.hide_version(1, expected_status=404)
        self.unhide_version(1, expected_status=404)

    def test_hide_last(self, moderator):
        self.login_user(moderator)

        doc_v1 = self.create_document(data="v1").json["document"]
        doc_v2 = self.modify_document(doc_v1, data="v2").json["document"]
        document_id = doc_v1["id"]

        # state 1: only V2 is hidden (it's the last one)
        self.hide_version(doc_v2)

        # for modo/admins
        self.get_document(document_id, data_should_be_present=True, version_should_be=doc_v1)
        self.get_version(doc_v1, data_should_be_present=True)
        self.get_version(doc_v2, data_should_be_present=True)

        # for other users
        self.logout_user()
        self.get_document(document_id, data_should_be_present=True, version_should_be=doc_v1)
        self.get_version(doc_v1, data_should_be_present=True)
        self.get_version(doc_v2, data_should_be_present=False)

        ## state 2 : V1 and V2 are hidden
        self.login_user(moderator)
        self.hide_version(doc_v1, expected_status=400)  # Can't hide all versions

        # # for modo/admins
        # self.get_document(document_id, data_should_be_present=False)
        # self.get_version(doc_v1, data_should_be_present=True)
        # self.get_version(doc_v2, data_should_be_present=True)

        # # for other users
        # self.logout_user()
        # self.get_document(document_id, data_should_be_present=False)
        # self.get_version(doc_v1, data_should_be_present=False)
        # self.get_version(doc_v2, data_should_be_present=False)



from tests.unit_tests.utils import BaseTest


class Test_HealthCheck(BaseTest):
    def test_main(self):
        response = self.get("/healthcheck")
        assert response.status_code == 200
        assert response.json["status"] == "ok"



from flask import request
from werkzeug.exceptions import NotFound, BadRequest

from flask_camp._schemas import schema
from flask_camp._utils import cook, current_api
from flask_camp.models._document import DocumentVersion, Document
from flask_camp._services._security import allow

rule = "/version/<int:version_id>"


@allow("anonymous", "authenticated", allow_blocked=True)
def get(version_id):
    """Get a given version of a document"""

    version = DocumentVersion.get(id=version_id)

    if version is None:
        raise NotFound()

    return {"status": "ok", "document": cook(version.as_dict())}


@allow("moderator")
@schema("modify_version.json")
def post(version_id):
    """Modify a version of a document. The only possible modification is hide/unhide a version"""
    version = DocumentVersion.get(id=version_id, with_for_update=True)

    if version is None:
        raise NotFound()

    old_last_version_id = version.document.last_version_id
    hidden = request.get_json()["hidden"]
    version.hidden = hidden
    current_api.database.session.flush()

    document = Document.get(id=version.document_id, with_for_update=True)
    document.update_last_version_id()

    needs_update = old_last_version_id != version.document.last_version_id

    if needs_update:
        current_api.before_document_save(document)

    current_api.add_log("hide_version" if hidden else "unhide_version", version=version, document=version.document)
    current_api.database.session.commit()

    if needs_update:
        version.document.clear_memory_cache()

    return {"status": "ok"}


@allow("admin")
@schema("action_with_comment.json")
def delete(version_id):
    """Delete a version of a document (only for admins)"""
    version = DocumentVersion.get(id=version_id, with_for_update=True)

    if version is None:
        raise NotFound()

    if DocumentVersion.query.filter_by(document_id=version.document_id).count() <= 1:
        raise BadRequest("Can't delete last version of a document")

    old_last_version_id = version.document.last_version_id
    document = Document.get(id=version.document_id, with_for_update=True)

    document.update_last_version_id(forbidden_id=version.id)
    needs_update = old_last_version_id != version.document.last_version_id

    if needs_update:
        current_api.before_document_save(document)

    current_api.database.session.delete(version)
    current_api.add_log("delete_version", version=version, document=version.document)
    current_api.database.session.commit()

    if needs_update:
        version.document.clear_memory_cache()

    return {"status": "ok"}



"""
flask_camp, Utils for flask_camp extension

Usage:
    flask_camp dev_env
    flask_camp init_db
    flask_camp add_admin <name> <email> <password>

Commands:
    dev_env     Start dev env, simply a redis on 6379 and a postgresql on 5432
    init_db     Init table in database
    add_admin   Adds an user with admin role
"""

import os
import subprocess

from flask import Flask
from docopt import docopt

from flask_camp import RestApi
from flask_camp.models import User


def _create_api():
    app = Flask(__name__)

    app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = False
    app.config["MAIL_DEFAULT_SENDER"] = app.config.get("MAIL_DEFAULT_SENDER", "do-not-reply@example.com")
    app.config["SQLALCHEMY_DATABASE_URI"] = app.config.get(
        "SQLALCHEMY_DATABASE_URI", "postgresql://flask_camp_user:flask_camp_user@localhost:5432/flask_camp"
    )

    app.config.from_prefixed_env()
    api = RestApi(app)

    return app, api


def main(args):

    if args["dev_env"]:  # pragma: no cover
        docker_file = os.path.join(os.path.dirname(__file__), "docker-compose.yml")
        subprocess.run(
            ["docker", "compose", "-f", docker_file, "up", "--remove-orphans", "--wait", "-d", "redis", "pg"],
            check=True,
        )

    elif args["init_db"]:
        app, api = _create_api()

        with app.app_context():
            api.database.create_all()

    elif args["add_admin"]:
        app, api = _create_api()

        with app.app_context():
            user = User(name=args["<name>"], roles=["admin"])
            user.set_password(args["<password>"])
            user.set_email(args["<email>"])
            user.validate_email(user._email_token)
            api.database.session.add(user)
            api.database.session.commit()

            print(f"{user} is created")


def main_entry_point():  # pragma: no cover
    main(docopt(__doc__))


if __name__ == "__main__":  # pragma: no cover
    main_entry_point()



# ## Password reset

# 1. as an anonymous user, the UI send a `POST /reset_password` with the mail
# 2. if the mail does not exists, or is not validated, the process stops (though, a normal response is sent)
# 3. A one-usage token is generated and sent to the user via a mail as a UI link like `/login?name=xx&token=yy`
# 4. The UI is responsible to do a `POST /login` with the username/token present in the request query
# 5. User is logged, and the token is removed.
# 6. UI is reponsible to show a password reset page
from datetime import timedelta, datetime
import re

from freezegun import freeze_time

from tests.unit_tests.utils import BaseTest


class Test_PasswordReset(BaseTest):
    def reset_password_and_get_token(self, user):
        with self.api.mail.record_messages() as outbox:
            r = self.reset_password(user._email)
            assert "expiration_date" in r.json
            token = re.sub(r"^(.*login_token=)", "", outbox[0].body)

        return token

    def test_simple(self, user):
        token = self.reset_password_and_get_token(user)

        self.login_user(user, expected_status=200)
        self.logout_user(expected_status=200)

        self.login_user(user, token=token, expected_status=200)

        r = self.get_user(user)
        assert r.json["user"]["email"] == user._email

        self.modify_user(user, token=token, new_password="a_new_pass")

        self.logout_user(expected_status=200)

        # test unique usage
        self.login_user(user, token=token, expected_status=401)

        # test that password has been updated
        self.login_user(user, password="password", expected_status=401)
        self.login_user(user, password="a_new_pass", expected_status=200)

    def test_email_not_found(self):
        with self.api.mail.record_messages() as outbox:
            self.reset_password("i_do@not_exists.fr", expected_status=200)
            assert len(outbox) == 0

    def test_user_is_not_validated(self, unvalidated_user):
        with self.api.mail.record_messages() as outbox:
            self.reset_password(unvalidated_user._email_to_validate, expected_status=200)
            assert len(outbox) == 0

    def test_bad_token(self, user):
        self.reset_password(user._email, expected_status=200)
        self.login_user(user, token="not the token", expected_status=401)

    def test_several_request(self, user):
        token_1 = self.reset_password_and_get_token(user)
        token_2 = self.reset_password_and_get_token(user)

        assert token_1 is not None
        assert token_2 is not None
        assert token_1 != token_2

        self.login_user(user, token=token_1, expected_status=401)
        self.login_user(user, token=token_2, expected_status=200)

    def test_expiration(self, user):

        token = self.reset_password_and_get_token(user)

        with freeze_time(datetime.now() + timedelta(days=3)):
            self.login_user(user, token=token, expected_status=401)
